import mxnet as mx

from .base import KernelFunction
from ..constants import INITIAL_NOISE_VARIANCE, NOISE_VARIANCE_LOWER_BOUND, NOISE_VARIANCE_UPPER_BOUND, DATA_TYPE, DEFAULT_ENCODING
from ..gluon_blocks_helpers import unwrap_parameter, IdentityScalarEncoding
from ..utils import create_encoding

__all__ = ['Coregionalization']


class Coregionalization(KernelFunction):
    """
    k(i, j) = K_{ij}, where K = W W^T + diag(rho).
    """
    def __init__(self, num_outputs, num_factors=16,
                 rho_init=INITIAL_NOISE_VARIANCE,
                 encoding_type=DEFAULT_ENCODING, **kwargs):

        super(Coregionalization, self).__init__(dimension=1, **kwargs)

        self.encoding_W_flat = IdentityScalarEncoding(
            dimension=num_outputs * num_factors)
        self.encoding_rho = create_encoding(encoding_type, rho_init,
                                            NOISE_VARIANCE_LOWER_BOUND,
                                            NOISE_VARIANCE_UPPER_BOUND,
                                            dimension=1)

        self.num_outputs = num_outputs
        self.num_factors = num_factors

        with self.name_scope():
            self.W_flat_internal = self.params.get(
                "W_internal", shape=(num_outputs * num_factors,),
                init=mx.init.Normal(),  # TODO: Use Xavier initialization here
                dtype=DATA_TYPE)
            self.rho_internal = self.params.get(
                "rho_internal", shape=(1,),
                init=mx.init.Constant(self.encoding_rho.init_val_int),
                dtype=DATA_TYPE)

    @staticmethod
    def _meshgrid(F, a, b):
        """
        Return coordinate matrices from coordinate vectors.

        Like https://docs.scipy.org/doc/numpy/reference/generated/numpy.meshgrid.html
        (with Cartesian indexing), but only supports two coordinate vectors as input.

        :param a: 1-D array representing the coordinates of a grid (length n) 
        :param b: 1-D array representing the coordinates of a grid (length m) 
        :return: coordinate matrix. 3-D array of shape (2, m, n).
        """
        aa = F.broadcast_mul(F.ones_like(F.expand_dims(a, axis=-1)), b)
        bb = F.broadcast_mul(F.ones_like(F.expand_dims(b, axis=-1)), a)
        return F.stack(bb, F.transpose(aa), axis=0)

    def _compute_gram_matrix(self, F, W_flat, rho):
        W = F.reshape(W_flat, shape=(self.num_outputs, self.num_factors))
        rho_vec = F.broadcast_mul(rho, F.ones(self.num_outputs, dtype=DATA_TYPE))
        return F.linalg.syrk(W) + F.diag(rho_vec)

    def hybrid_forward(self, F, ind1, ind2, W_flat_internal, rho_internal):
        W_flat = self.encoding_W_flat.get(F, W_flat_internal)
        rho = self.encoding_rho.get(F, rho_internal)
        K = self._compute_gram_matrix(F, W_flat, rho)
        ind1 = self._check_input_shape(F, ind1)
        if ind2 is not ind1:
            ind2 = self._check_input_shape(F, ind2)
        ind = self._meshgrid(F, ind1, ind2)
        return F.transpose(F.squeeze(F.gather_nd(K, ind)))

    def diagonal(self, F, ind):
        ind = self._check_input_shape(F, ind)
        W_flat = self.encoding_W_flat.get(F, unwrap_parameter(F, self.W_flat_internal, ind))
        rho = self.encoding_rho.get(F, unwrap_parameter(F, self.rho_internal, ind))
        K = self._compute_gram_matrix(F, W_flat, rho)
        K_diag = F.diag(K)
        return F.take(K_diag, ind)

    def diagonal_depends_on_X(self):
        return True

    def param_encoding_pairs(self):
        return [
            (self.W_flat_internal, self.encoding_W_flat),
            (self.rho_internal, self.encoding_rho),
        ]



import mxnet as mx
import logging
import time

from .utils import param_to_pretty_string

logger = logging.getLogger(__name__)


class DebugGPRegression(object):
    """
    Supports finding errors in GaussianProcessRegression.fit. For each
    criterion evaluation, we store the input arguments before and the
    criterion value afterwards. Storage is done to a rolling sequence
    of local files.
    """
    def __init__(self, fname_msk='debug_gpr_{}', rolling_size=3):
        self.fname_msk = fname_msk
        self.rolling_size = rolling_size
        self.global_counter = 0
        self.optim_counter = -1
        self.local_counter = 0

    def start_optimization(self):
        self.optim_counter += 1
        self.local_counter = 0

    def store_args(self, params, X, Y, param_encoding_pairs):
        arg_dict = {
            'features': X,
            'targets': Y}
        for param in params:
            arg_dict[param.name] = param.data(ctx=mx.cpu())
        fname = self._filename()
        mx.nd.save(fname + '_args.nd', arg_dict)
        with open(fname + '_args.txt', 'w') as f:
            self._write_meta(f)
            for param, encoding in param_encoding_pairs:
                f.write(param_to_pretty_string(param, encoding) + '\n')

    def store_value(self, value):
        fname = self._filename()
        with open(fname + '_value.txt', 'w') as f:
            f.write('value = {}\n'.format(value))
            self._write_meta(f)
        # Advance counters
        self.global_counter += 1
        self.local_counter += 1

    def _filename(self):
        return self.fname_msk.format(self.global_counter % self.rolling_size)

    def _write_meta(self, f):
        f.write('optim_counter = {}\n'.format(self.optim_counter))
        f.write('local_counter = {}\n'.format(self.local_counter))
        f.write('global_counter = {}\n'.format(self.global_counter))
        f.write('time = {}\n'.format(time.time()))



from typing import Tuple, Union
import ConfigSpace as CS
import ConfigSpace.hyperparameters as CSH
import copy

from .hp_ranges import HyperparameterRanges_CS

RESOURCE_ATTR_PREFIX = 'RESOURCE_ATTR_'


class ExtendedConfiguration(object):
    """
    This class facilitates handling extended configs, which consist of a normal
    config and a resource attribute.

    The config space hp_ranges is extended by an additional resource
    attribute. Note that this is not a hyperparameter we optimize over,
    but it is under the control of the scheduler.
    Its allowed range is [1, resource_attr_range[1]], which can be larger than
    [resource_attr_range[0], resource_attr_range[1]]. This is because extended
    configs with resource values outside of resource_attr_range may arise (for
    example, in the early stopping context, we may receive data from
    epoch < resource_attr_range[0]).

    """
    def __init__(
            self, hp_ranges: HyperparameterRanges_CS, resource_attr_key: str,
            resource_attr_range: Tuple[int, int]):
        assert resource_attr_range[0] >= 1
        assert resource_attr_range[1] >= resource_attr_range[0]
        self.hp_ranges = hp_ranges
        self.resource_attr_key = resource_attr_key
        self.resource_attr_range = resource_attr_range
        # Extended configuration space including resource attribute
        config_space_ext = copy.deepcopy(hp_ranges.config_space)
        self.resource_attr_name = RESOURCE_ATTR_PREFIX + resource_attr_key
        # Allowed range: [1, resource_attr_range[1]]
        config_space_ext.add_hyperparameter(CSH.UniformIntegerHyperparameter(
            name=self.resource_attr_name, lower=1,
            upper=resource_attr_range[1]))
        self.hp_ranges_ext = HyperparameterRanges_CS(
            config_space_ext, name_last_pos=self.resource_attr_name)

    def get(self, config: CS.Configuration, resource: int) -> CS.Configuration:
        """
        Create extended config with resource added.

        :param config:
        :param resource:
        :return: Extended config
        """
        values = copy.deepcopy(config.get_dictionary())
        values[self.resource_attr_name] = resource
        return CS.Configuration(self.hp_ranges_ext.config_space, values=values)

    def remap_resource(
            self, config_ext: CS.Configuration, resource: int,
            as_dict: bool=False) -> Union[CS.Configuration, dict]:
        x_dct = copy.copy(config_ext.get_dictionary())
        x_dct[self.resource_attr_name] = resource
        if as_dict:
            return x_dct
        else:
            return CS.Configuration(
                self.hp_ranges_ext.config_space, values=x_dct)

    def remove_resource(
            self, config_ext: CS.Configuration,
            as_dict: bool=False) -> Union[CS.Configuration, dict]:
        x_dct = copy.copy(config_ext.get_dictionary())
        del x_dct[self.resource_attr_name]
        if as_dict:
            return x_dct
        else:
            return CS.Configuration(self.hp_ranges.config_space, values=x_dct)

    def from_dict(self, config_dct: dict) -> CS.Configuration:
        # Note: Here, the key for resource is resource_attr_key, not
        # resource_attr_name
        return CS.Configuration(
            self.hp_ranges_ext if self.resource_attr_key in config_dct \
                else self.hp_ranges, values=config_dct)



# coding: utf-8
# pylint: disable=wrong-import-position
""" AutoGluon: AutoML Toolkit for Deep Learning """
from __future__ import absolute_import
from .version import __version__

import logging
logger = logging.getLogger("distributed.utils_perf")
logger.setLevel(logging.ERROR)

logger = logging.getLogger("distributed.logging.distributed")
logger.setLevel(logging.ERROR)

logger = logging.getLogger("distributed.worker")
logger.setLevel(logging.ERROR)

from .utils.try_import import *
try_import_mxnet()

from . import scheduler, searcher, utils
from .scheduler import get_cpu_count, get_gpu_count

from .utils import *
from .core import *
from .task import *



import logging

from .tabular_nn_dataset import TabularNNDataset
from ..abstract import model_trial
from ....utils.exceptions import TimeLimitExceeded
from ......core import args

logger = logging.getLogger(__name__)


@args()
def tabular_nn_trial(args, reporter):
    """ Training and evaluation function used during a single trial of HPO """
    try:
        model, args, util_args = model_trial.prepare_inputs(args=args)

        train_dataset = TabularNNDataset.load(util_args.train_path)
        val_dataset = TabularNNDataset.load(util_args.val_path)
        y_val = val_dataset.get_labels()

        fit_model_args = dict(X_train=train_dataset, y_train=None, X_val=val_dataset)
        predict_proba_args = dict(X=val_dataset)
        model_trial.fit_and_save_model(model=model, params=args, fit_args=fit_model_args, predict_proba_args=predict_proba_args, y_val=y_val,
                                       time_start=util_args.time_start, time_limit=util_args.get('time_limit', None), reporter=reporter)
    except Exception as e:
        if not isinstance(e, TimeLimitExceeded):
            logger.exception(e, exc_info=True)
        reporter.terminate()



import logging

from pandas import DataFrame

from .abstract import AbstractFeatureGenerator

logger = logging.getLogger(__name__)


class IdentityFeatureGenerator(AbstractFeatureGenerator):
    """IdentityFeatureGenerator simply passes the data along without alterations."""
    def _fit_transform(self, X: DataFrame, **kwargs) -> (DataFrame, dict):
        X_out = self._transform(X)
        return X_out, self.feature_metadata_in.type_group_map_special

    def _transform(self, X: DataFrame) -> DataFrame:
        return X

    @staticmethod
    def get_default_infer_features_in_args() -> dict:
        return dict()
